// Copyright 2023 Georges Racinet <georges.racinet@octobus.net>
//
// This software may be used and distributed according to the terms of the
// GNU General Public License version 2 or any later version.
// SPDX-License-Identifier: GPL-2.0-or-later
use std::env;
use std::error::Error;
use std::ffi::OsString;
use std::net::SocketAddr;
use std::path::PathBuf;

#[derive(Debug)]
#[allow(clippy::upper_case_acronyms)]
/// Necessary infortion to bind the listening socket
pub enum BindAddress {
    Unix(PathBuf),
    TCP(SocketAddr),
}

#[derive(Debug, Clone)]
#[allow(clippy::upper_case_acronyms)]
/// Necessary infortion to call a server (e.g., HGitaly (Python) sidecar)
pub enum ServerAddress {
    Unix(PathBuf),
    URI(String),
}

/// Parse the URL to listen to
///
/// There are two cases: `unix:/some/path/sock` and `tcp://hostname:port`, where
/// hostname can only be an IP address (v4 or v6).
///
/// This is a toy implementation anyway, we'll add a proper dependency to the `url` crate later on
fn parse_listen_url(url: &str) -> Result<BindAddress, Box<dyn Error>> {
    if let Some(path) = url.strip_prefix("unix:") {
        return Ok(BindAddress::Unix(path.into()));
    }
    if let Some(addr) = url.strip_prefix("tcp://") {
        return Ok(BindAddress::TCP(addr.parse()?));
    }
    Err("Unsupported URL".into())
}

/// Parse the URI of a gRPC server
///
/// There are two cases: `unix:/some/path/sock` and URIs directly supported by Tonic
/// (`http` or `https` schemes). The latter are not validated at this point.
fn parse_server_uri(uri: &str) -> Result<ServerAddress, Box<dyn Error>> {
    if let Some(path) = uri.strip_prefix("unix:") {
        return Ok(ServerAddress::Unix(path.into()));
    }
    Ok(ServerAddress::URI(uri.to_owned()))
}

#[derive(Debug)]
pub struct Config {
    pub repositories_root: PathBuf,
    pub listen_address: BindAddress,
    pub hg_executable: OsString,
    pub hgitaly_sidecar_address: ServerAddress,
}

impl Config {
    /// Read configuration from environment variables
    ///
    /// This is useful for Gitaly Comparison tests, especially until we can
    /// support the HGitaly configuration file (i.e., HGRC)
    pub fn from_env() -> Self {
        Config {
            repositories_root: env::var_os("RHGITALY_REPOSITORIES_ROOT")
                .expect("RHGITALY_REPOSITORIES_ROOT not set in environment")
                .into(),
            listen_address: parse_listen_url(
                &env::var("RHGITALY_LISTEN_URL")
                    .expect("RHGITALY_LISTEN_URL not set in environment"),
            )
            .expect("Could not parse listen URL"),
            hgitaly_sidecar_address: parse_server_uri(
                &env::var("RHGITALY_SIDECAR_ADDRESS")
                    .expect("RHGITALY_SIDECAR_ADDRESS not set in environment"),
            )
            .expect("Could not parse sidecar URL"),
            hg_executable: env::var("RHGITALY_HG_EXECUTABLE").map_or_else(
                |e| match e {
                    env::VarError::NotPresent => "hg".into(),
                    env::VarError::NotUnicode(os_string) => os_string,
                },
                |v| v.into(),
            ),
        }
    }
}
