// Copyright 2023 Georges Racinet <georges.racinet@octobus.net>
//
// This software may be used and distributed according to the terms of the
// GNU General Public License version 2 or any later version.
// SPDX-License-Identifier: GPL-2.0-or-later
use std::ffi::{OsStr, OsString};
use std::fmt;
use std::marker::Send;
use std::os::unix::ffi::OsStringExt;
use std::path::{Path, PathBuf};
use std::process::Stdio;
use std::sync::Arc;

use tokio::fs;
use tokio::io::{AsyncBufReadExt, BufReader};
use tokio::process::Command;
use tokio::select;
use tokio::sync::mpsc::{self, Sender};
use tokio_stream::wrappers::ReceiverStream;
use tokio_util::sync::CancellationToken;
use tonic::metadata::MetadataMap;
use tonic::{Response, Status};
use tracing::{debug, info, span::Span, warn};

use hg::changelog::Changelog;
use hg::config::Config as CoreConfig;
use hg::repo::{Repo, RepoError};

use super::config::Config;
use super::gitaly::{Repository, User};
use super::metadata::{
    get_boolean_md_value, ACCEPT_MR_IID_KEY, HG_GIT_MIRRORING_MD_KEY, NATIVE_PROJECT_MD_KEY,
    SKIP_HOOKS_MD_KEY,
};
use super::process;
use super::streaming::BlockingResponseSender;

/// Represent errors that are due to a wrong repository specification.
///
/// In terms of gRPC methods, the specifications is usually enclosed in  a [`Repository`] message
/// and these errors are considered to be client errors.
#[derive(Debug, PartialEq, Eq)]
pub enum RepoSpecError {
    MissingSpecification,
    UnknownStorage(String),
    RepoNotFound(PathBuf),
}

/// Represent errors loading a repository (bad specification or internal errors)
#[derive(Debug, derive_more::From)] // TODO add PartialEq, but do it in core for RepoError first
pub enum RepoLoadError {
    #[from]
    SpecError(RepoSpecError),
    LoadError(RepoError),
}

impl From<RepoError> for RepoLoadError {
    fn from(value: RepoError) -> Self {
        if let RepoError::NotFound { at } = value {
            return Self::SpecError(RepoSpecError::RepoNotFound(at));
        }
        Self::LoadError(value)
    }
}

/// Default conversion of ['RepoSpecError'] into a gRPC ['Status']
///
/// This function does not care to precisely match the error details, focusing on the error
/// codes instead.
///
/// The resulting codes match the most common behaviour of Gitaly, which actually behaves more
/// and more like this with time (e.g., as internal Git error get catched and handled).
pub fn default_repo_spec_error_status(err: RepoSpecError) -> Status {
    match err {
        RepoSpecError::MissingSpecification => Status::invalid_argument("repository not set"),
        RepoSpecError::UnknownStorage(storage) => Status::invalid_argument(format!(
            "GetStorageByName: no such storage: \"{}\"",
            storage
        )),
        RepoSpecError::RepoNotFound(at) => {
            Status::not_found(format!("repository at \"{}\" not found", at.display()))
        }
    }
}

pub fn repo_path(config: &Config, repo: &Repository) -> Result<PathBuf, RepoSpecError> {
    if repo.storage_name != "default" {
        return Err(RepoSpecError::UnknownStorage(repo.storage_name.clone()));
    }
    let root = &config.repositories_root;

    let res = match repo.relative_path.strip_suffix(".git") {
        Some(stripped) => root.join(stripped.to_owned() + ".hg"),
        None => root.join(&repo.relative_path),
    };

    // TODO forbid climbing up (same in Python, will be necessary for HGitaly3, even
    // though clients are deeply trusted.
    Ok(res)
}

/// Default gRPC error ['Status'] for repository not found.
///
/// To be used if repository path does not exist on disk.
pub fn default_repo_not_found_error_status(path: &Path) -> Status {
    Status::not_found(format!(
        "Mercurial repository at {} not found",
        path.display()
    ))
}

pub async fn checked_repo_path<'a>(
    config: &Config,
    gl_repo: Option<&'a Repository>,
) -> Result<(&'a Repository, PathBuf), RepoSpecError> {
    let repo = gl_repo
        .as_ref()
        .ok_or(RepoSpecError::MissingSpecification)?;
    let path = repo_path(config, repo)?;
    if match fs::metadata(&path).await {
        Ok(md) => md.is_dir(),
        Err(_) => false,
    } {
        return Ok((repo, path));
    }
    Err(RepoSpecError::RepoNotFound(path))
}

/// Return a path to virtual filesystem for the repository store.
///
/// As of this writing, this is nothing but a [`Path`], but it could become something
/// more advanced in the future (perhaps not as much as `hg_core` `Vfs` type, though).
///
/// Parameter `repo` is an `Option`, so that a service method can pass directly
/// something like `&request.repository`, with `None` giving rise to the natural
/// `MissingSpecification` error.
///
/// If the repository is not found on disc, the appropriate error is also returned.
pub async fn repo_store_vfs(
    config: &Config,
    repo: &Option<Repository>,
) -> Result<PathBuf, RepoSpecError> {
    let root = checked_repo_path(config, repo.as_ref()).await?.1;
    Ok(root.join(".hg/store"))
}

/// Read repository VFS requirements and tell whether one fulfills the given condition
///
/// Note that some requirements are in the 'store` VFS (most of those that
/// will be of interest with RHGitaly), while some others are on the 'working dir' VFS.
///
/// Inspired by [`hg::requirements::load`] with thw following differences:
///
/// - we do not need to build a full `HashSet`, just to check a single value. If the caller
///   needs to use the requirements repeatedly, we will provide a more general parsing function
///   with storable results.
/// - this is async
pub async fn has_requirement_with(
    vfs: &Path,
    condition: impl FnMut(&[u8]) -> bool,
) -> Result<bool, Status> {
    let path = vfs.join("requires"); // TODO const
    Ok(fs::read(&path)
        .await
        .map_err(|e| {
            Status::internal(format!(
                "Could not open requirements at {}: {}",
                path.display(),
                e
            ))
        })?
        .split(|b| *b == b'\n')
        .any(condition))
}

/// Return `Status::Unimplemented` if the repo has the `largefiles` requirement.
pub async fn unimplemented_if_largefiles(
    config: &Config,
    repo: &Option<Repository>,
) -> Result<(), Status> {
    let store_vfs = repo_store_vfs(config, repo)
        .await
        .map_err(default_repo_spec_error_status)?;
    if has_requirement_with(&store_vfs, |req| req == b"largefiles").await? {
        // TODO const
        return Err(Status::unimplemented("largefiles repo requirement"));
    };
    Ok(())
}

pub fn load_repo(config: &Config, opt_repo: Option<&Repository>) -> Result<Repo, RepoLoadError> {
    let repo = opt_repo.ok_or(RepoSpecError::MissingSpecification)?;
    // TODO load core config once and for all and store it in our Config.
    // but do we really want to have all this unrelated stuff around?
    let core_config =
        CoreConfig::load_non_repo().expect("Should have been able to read Mercurial config");
    // TODO better to use Repo::new_at_path, but it is private
    // (Repo::find does verifications that are useless to us.
    // At least it does not try to climb up when passed an explicit path)
    Ok(Repo::find(&core_config, Some(repo_path(config, repo)?))?)
}

/// Trait for requests with a repository field
///
/// It provides the needed uniformity for methods such as [`load_repo_and_stream`]
pub trait RequestWithRepo: Send + 'static {
    /// Grab a reference to the [`Repository`] field from the request.
    ///
    /// Like all submessages, the repository is optional.
    fn repository_ref(&self) -> Option<&Repository>;
}

/// Trait for requests whose treatment involves spawning a hg child process
///
/// It provides the needed uniformity for methods such as [`load_repo_and_spawn_hg`]
pub trait RequestHgSpawnable: RequestWithRepo {
    /// Grab a reference to the [`User`] field from the request.
    ///
    /// Like all submessages, the user is optional if it is indeed part of protocol.
    ///
    /// In the case of read-only gRPC methods, it is totally acceptable not to have
    /// any [`User`] field, whence the blanket implementation returning `None`.
    fn user_ref(&self) -> Option<&User> {
        None
    }
}

/// Load a repository and initiate streaming responses
///
/// This setups the `mpsc` channel expected by Tonic and spawns a blocking task (typically run
/// in a separate thread) loads the repository, and passes over the repository and the transmitting
/// end of the channel to the caller supplied closure.
///
/// The `and_then` closure must perform its streaming by sending `Result<Resp, Status>` values
/// on the channel, using the provided [`BlockingResponseSender`].
///
/// If the repository loading fails, the appropriate gRPC error response is sent over
/// or logged if sending is not possible.
///
/// Because Gitaly's error responses are not uniform, and we want to match them closely,
/// ethe caller has to supply a callable for conversion of [`RepoSpecError`] to the appropriate
/// [`Status`]. The [`default_repo_spec_error_status`] function can be used in the majority of
/// cases and serve as an example for other cases.
pub fn load_repo_and_stream<Req: RequestWithRepo, Resp: fmt::Debug + Send + 'static>(
    config: Arc<Config>,
    request: Req,
    repo_spec_error_status: impl Fn(RepoSpecError) -> Status + Send + 'static,
    and_then: impl FnOnce(Req, Repo, BlockingResponseSender<Resp>) + Send + 'static,
) -> Result<Response<ReceiverStream<Result<Resp, Status>>>, Status> {
    // no point having channel capacity for several messages, since `blocking_send` is the only
    // way to use it.
    let (tx, rx) = mpsc::channel(1);
    let current_span = Span::current();
    tokio::task::spawn_blocking(move || {
        let btx: BlockingResponseSender<Resp> = tx.into();
        let _entered = current_span.enter();
        match load_repo(&config, request.repository_ref()) {
            Err(RepoLoadError::SpecError(e)) => btx.send(Err(repo_spec_error_status(e))),
            Err(RepoLoadError::LoadError(e)) => btx.send(Err(Status::internal(format!(
                "Error loading repository: {:?}",
                e
            )))),
            Ok(repo) => and_then(request, repo, btx),
        }
    });
    Ok(Response::new(ReceiverStream::new(rx)))
}

/// Load a repository in a separate thread and hand it over to a closure
///
/// This creates a new thread suitable for blocking operations, using
/// [`tokio::task::spawn_blocking`] and then hands it over together with the orignal request
/// to the `and_then` closure, whose return value is finally returned to the caller.
/// The closure is at liberty to use any blocking operation (most calls to `hg-core` are blocking).
///
/// If the repository loading fails, the appropriate gRPC error [`Status`] is returned
///
/// `repo_spec_error_status` plays the same role as in [`load_repo_and_stream`], and again
/// the [`default_repo_spec_error_status`] function can be used in the majority of
/// cases.
pub async fn load_repo_and_then<Req: RequestWithRepo, Res: Send + 'static>(
    config: Arc<Config>,
    request: Req,
    repo_spec_error_status: impl Fn(RepoSpecError) -> Status + Send + 'static,
    and_then: impl FnOnce(Req, Repo) -> Result<Res, Status> + Send + 'static,
) -> Result<Res, Status> {
    let current_span = Span::current();
    tokio::task::spawn_blocking(move || {
        let _entered = current_span.enter();
        match load_repo(&config, request.repository_ref()) {
            Err(RepoLoadError::SpecError(e)) => Err(repo_spec_error_status(e)),
            Err(RepoLoadError::LoadError(e)) => Err(Status::internal(format!(
                "Error loading repository: {:?}",
                e
            ))),
            Ok(repo) => and_then(request, repo),
        }
    })
    .await
    .map_err(|join_err| {
        if join_err.is_cancelled() {
            // According to https://grpc.io/docs/guides/error/, it should be
            // `Unavailable()` if this is a graceful shutdown, but now the question is how
            // to tell that apart from user cancellation.
            Status::cancelled("Inner blocking task on Mercurial repo was cancelled")
        } else {
            Status::internal(format!("Unexpected error in inner thread: {:?}", join_err))
        }
    })?
}

/// Load the repository and expose it to the caller, and then spawn a `hg` child process on it.
///
/// First repository loading is similar to [`load_repo_and_then`], executing the provided
/// `before_spawn` closure on the repository and the prepared [`Command`], allowing for
/// tuning according to configuration, perform resolutions that could be faster or more
/// practical from Rust code, and early checks. Errors returned by this closure will be sent
/// back to the caller immediately, hence avoiding to spwan the child process.
///
/// If the request specifies an [`User`], all necessary environment variables are given to
/// the child process so that mutation on behalf of the given user can work. This is very
/// similar to the legacy code in the Rails application. A difference lies in the way the (usually
/// necessary) `HGRCPATH` environment variable is given to the child process: nothing special is
/// done about it, hence it is assumed it is set on the whole RHGitaly service.
///
/// If the optional `stdout_tx` is given, the child process standard output will be captured and
/// sent line by line over it.
/// In general, other means of obtaining information should be preferred, but there is
/// sometimes nothing else that can be used.
///
/// The entire environment of the current process is passed down to the child process.
///
/// The path to the `hg` executable is taken from `Config`.
// This function does have too many arguments, but we will refactor
// when more use-cases are available.
#[allow(clippy::too_many_arguments)]
pub async fn load_repo_and_spawn_hg<Req: RequestHgSpawnable, Args, Arg>(
    config: Arc<Config>,
    request: Req,
    metadata: &MetadataMap,
    args: Args,
    repo_spec_error_status: impl Fn(RepoSpecError) -> Status + Send + 'static + Copy,
    shutdown_token: &CancellationToken,
    stdout_tx: Option<Sender<String>>,
    before_spawn: impl FnOnce(Req, Repo, &mut Command) -> Result<(), Status> + Send + 'static,
) -> Result<i32, Status>
where
    Args: IntoIterator<Item = Arg>,
    Arg: AsRef<OsStr>,
{
    let (gitaly_repo, repo_path) = checked_repo_path(&config, request.repository_ref())
        .await
        .map_err(repo_spec_error_status)?;
    let mut cmd = Command::new(&config.hg_executable);
    cmd.arg("-R")
        .arg(repo_path)
        .args(args)
        .env("GL_REPOSITORY", &gitaly_repo.gl_repository)
        .env("GL_PROTOCOL", "web"); // same hardcoding as in Gitaly

    if let Some(user) = request.user_ref() {
        let mirroring = get_boolean_md_value(metadata, HG_GIT_MIRRORING_MD_KEY, false);
        let native = get_boolean_md_value(metadata, NATIVE_PROJECT_MD_KEY, false);
        let skip_gl_hooks = get_boolean_md_value(metadata, SKIP_HOOKS_MD_KEY, false);

        if native {
            cmd.arg("--config").arg("heptapod.native=true");
        }
        if !mirroring {
            cmd.arg("--config").arg("heptapod.no-git=true");
        }
        if skip_gl_hooks {
            cmd.env("'HEPTAPOD_SKIP_ALL_GITLAB_HOOKS'", "yes");
        }

        cmd.env("HEPTAPOD_USERINFO_GL_ID", &user.gl_id);
        cmd.env("HEPTAPOD_USERINFO_USERNAME", &user.gl_username);
        cmd.env(
            "HEPTAPOD_USERINFO_NAME",
            &OsString::from_vec(user.name.clone()),
        );
        cmd.env(
            "HEPTAPOD_USERINFO_EMAIL",
            &OsString::from_vec(user.email.clone()),
        );

        if let Some(v) = metadata.get(ACCEPT_MR_IID_KEY) {
            if let Ok(iid) = v.to_str() {
                cmd.env("HEPTAPOD_ACCEPT_MR_IID", iid);
            }
        }
    };
    let mut cmd = load_repo_and_then(config, request, repo_spec_error_status, move |req, repo| {
        before_spawn(req, repo, &mut cmd)?;
        Ok(cmd)
    })
    .await?;

    if stdout_tx.is_some() {
        cmd.stdout(Stdio::piped());
    }
    debug!("Spawning command {:#?}", cmd);
    let shutdown_token = shutdown_token.clone();
    let mut hg = cmd
        .spawn()
        .map_err(|e| Status::internal(format!("Error spawning Mercurial subprocess: {}", e)))?;

    let stdout_and_tx = stdout_tx.map(|stdtx| {
        (
            hg.stdout
                .take()
                .expect("Spawned process has no stdout (already been taken?)"),
            stdtx,
        )
    });

    let token = CancellationToken::new();
    let _drop_guard = token.clone().drop_guard();

    let (tx, mut rx) = mpsc::channel(1);
    tokio::spawn(async move {
        let hg_status = select! {
            res = hg.wait() => res.map_err(|e| {
                Status::internal(
                    format!("Error waiting for Mercurial subprocess: {}", e))
                }),
            _ = token.cancelled() => {
                // TODO logs from a subthread are not in the context of the request
                // hence will turn pretty useless without correlation_id etc.
                info!("Task cancelled, terminating Mercurial child process with SIGTERM");
                process::terminate(hg).await;
                Err(Status::cancelled("Task dropped, probably due to \
                                       client-side cancellation"))
            },
            _ = shutdown_token.cancelled() => {
                warn!("General shutdown required, terminating Mercurial child \
                       process with SIGTERM");
                process::terminate(hg).await;
                Err(Status::unavailable("RHGitaly server is shutting down"))
            },
        };
        tx.send(hg_status).await
    });

    let hg_status = if let Some((stdout, stdout_tx)) = stdout_and_tx {
        let mut reader = BufReader::new(stdout).lines();

        async move {
            let hg_status: Result<_, Status>;
            loop {
                select! {
                    maybe_line = reader.next_line() => {
                        match maybe_line {
                            Ok(Some(line)) => {
                                info!("Line: {}", line);
                                if stdout_tx.send(line).await.is_err() {
                                    // we have no other choice than ignoring it,
                                    // although it is probably symptom
                                    // of some really unexpected problem
                                    warn!("Subprocess stdout receiver already dropped!")
                                }
                            },
                            Err(e) => {
                                // probably the error is not due to stdout already closed, but
                                // in any case, we must let the other arm of `select!` run so
                                // that the process is eventually reaped
                                warn!("Got error reading from child process stdout: {}", e)
                            }
                            Ok(None) => {},
                        }
                    },
                    res = rx.recv() => {
                        hg_status = res.unwrap_or_else(|| Err(Status::internal(
                            "Channel closed before sending back Mercurial subprocess status")));
                        break;
                    }
                }
            }
            hg_status
        }
        .await?
    } else {
        rx.recv().await.unwrap_or_else(|| {
            Err(Status::internal(
                "Channel closed before sending back Mercurial subprocess status",
            ))
        })?
    };

    hg_status.code().ok_or(Status::internal(
        "Mercurial subprocess killed or stopped by signal",
    ))
}

/// Load a repository and its changelog in a separate thread and hand it over to a closure
///
/// See [`load_repo_and_then`] on which this builds upon for more details.
pub async fn load_changelog_and_then<Req: RequestWithRepo, Res: Send + 'static>(
    config: Arc<Config>,
    request: Req,
    repo_spec_error_status: impl Fn(RepoSpecError) -> Status + Send + 'static,
    and_then: impl FnOnce(Req, &Repo, &Changelog) -> Result<Res, Status> + Send + 'static,
) -> Result<Res, Status> {
    load_repo_and_then(config, request, repo_spec_error_status, |req, repo| {
        let cl = repo
            .changelog()
            .map_err(|e| Status::internal(format!("Could not open changelog: {:?}", e)))?;
        and_then(req, &repo, &cl)
    })
    .await
}

/// Load a repository and its changelog in a separate thread for streaming responses
///
/// See [`load_repo_and_stream`] on which this builds upon for more details.
pub fn load_changelog_and_stream<Req: RequestWithRepo, Resp: fmt::Debug + Send + 'static>(
    config: Arc<Config>,
    request: Req,
    repo_spec_error_status: impl Fn(RepoSpecError) -> Status + Send + 'static,
    and_then: impl FnOnce(Req, &Repo, &Changelog, BlockingResponseSender<Resp>) + Send + 'static,
) -> Result<Response<ReceiverStream<Result<Resp, Status>>>, Status> {
    load_repo_and_stream(
        config,
        request,
        repo_spec_error_status,
        |req, repo, tx| match repo.changelog() {
            Ok(cl) => and_then(req, &repo, &cl, tx),
            Err(e) => tx.send(Err(Status::internal(format!(
                "Could not open changelog: {:?}",
                e
            )))),
        },
    )
}

#[cfg(test)]
mod tests {

    use super::*;
    use crate::config::{BindAddress, ServerAddress};
    use crate::gitaly::ServerInfoRequest;

    #[test]
    fn test_repo_path() {
        let config = Config {
            repositories_root: "/repos".into(),
            hg_executable: "hg".into(),
            listen_address: BindAddress::Unix("/tmp/rhgitaly.socket".into()),
            hgitaly_sidecar_address: ServerAddress::Unix("/tmp/hgitaly.socket".into()),
        };
        let mut repo = Repository::default();
        repo.storage_name = "default".into();
        repo.relative_path = "foo/bar".into();
        assert_eq!(repo_path(&config, &repo), Ok("/repos/foo/bar".into()));

        repo.relative_path = "foo/bar.git".into();
        assert_eq!(repo_path(&config, &repo), Ok("/repos/foo/bar.hg".into()));
    }

    impl RequestWithRepo for ServerInfoRequest {
        fn repository_ref(&self) -> Option<&Repository> {
            None // would not be acceptable in main code
        }
    }
    impl RequestHgSpawnable for ServerInfoRequest {}

    #[test]
    fn test_request_hg_spawnable() {
        assert!(ServerInfoRequest::default().user_ref().is_none());
    }
}
